# Ω*
#               ■          ■■■■■  
#               ■         ■■   ■■ 
#               ■        ■■     ■ 
#               ■        ■■       
#     ■■■■■     ■        ■■■      
#    ■■   ■■    ■         ■■■     
#   ■■     ■■   ■          ■■■■   
#   ■■     ■■   ■            ■■■■ 
#   ■■■■■■■■■   ■              ■■■
#   ■■          ■               ■■
#   ■■          ■               ■■
#   ■■     ■    ■        ■■     ■■
#    ■■   ■■    ■   ■■■  ■■■   ■■ 
#     ■■■■■     ■   ■■■    ■■■■■


"""
Moca System Core Module
モカシステムコアモジュール
茉客系统核心模块

[MocaCore]

Copyright (c) 2020.1.17 [el.ideal-ideas]

This software is released under the MIT License. see LICENSE.txt.

https://www.el-ideal-ideas.com
"""

# -- Imports --------------------------------------------------------------------------

from typing import Any, Optional, Iterable, Union, Callable, List
from threading import Thread
from multiprocessing import Process
from time import sleep
from pathlib import Path
from inspect import stack
from uuid import uuid4
from random import choice
from datetime import datetime
from os import _exit
from sys import exit
from io import StringIO
from json import dump, dumps
from .constants import DIGITS, HIRAGANA, KATAKANA, MOCHI_MOCHI, MENDAKO, ConsoleColor

# -------------------------------------------------------------------------- Imports --

# -- Public Functions --------------------------------------------------------------------------


def if_not_none(default: Any = None,
                *args) -> Any:
    """
    Return the first item that is not None.
    If all item in args is None, return default.
    """
    for item in args:
        if item is not None:
            return item
    return default


def if_not_false(default: Any = None,
                 *args) -> Any:
    """
    Return the first item that is not False.
    If all item in args is False, return default.
    """
    for item in args:
        if item:
            return item
    return default


def is_not_none(*args) -> bool:
    """If all item in args is not None, return True."""
    for item in args:
        if item is None:
            return False
    return True


def is_not_false(*args) -> bool:
    """If all item in args is True, return True."""
    for item in args:
        if not item:
            return False
    return True


def get_caller_name() -> str:
    """Return caller name."""
    try:
        return stack()[2][3]
    except IndexError:
        return 'Unknown'
    except Exception as error:  # Just to be sure
        return f'<Exception: {error}>'


def get_self_name() -> str:
    """Return self name."""
    try:
        return stack()[1][3]
    except IndexError:
        return 'Unknown'
    except Exception as error:  # Just to be sure
        return f'<Exception: {error}>'


def get_random_string(length: int,
                      characters: Optional[Iterable] = None) -> str:
    """
    Create a random string by characters.
    If characters is None create random string use uuid4.
    """
    if characters is None:
        return ''.join([uuid4().hex for _ in range((length // 32) + 1)])[0:length]
    else:
        return ''.join([str(choice(characters)) for _ in range(length)])


def get_random_string_by_digits(length: int) -> str:
    """Create a random string use digits."""
    return get_random_string(length, DIGITS)


def get_random_string_by_hiragana(length: int) -> str:
    """Create a random string use japanese hiragana."""
    return get_random_string(length, HIRAGANA)


def get_random_string_by_katakana(length: int) -> str:
    """Create a random string use japanese katakana."""
    return get_random_string(length, KATAKANA)


def get_random_string_by_kana(length: int) -> str:
    """Create a random string use japanese kana."""
    return get_random_string(length, KATAKANA + HIRAGANA)


def get_random_string_by_mochi(length: int) -> str:
    """Create a random string use mochi ٩(ˊᗜˋ*)و"""
    return get_random_string(length, MOCHI_MOCHI)


def get_random_string_by_tako(length: int) -> str:
    """Create a random string use tako ♫ヽ(゜∇゜ヽ)♪ """
    return get_random_string(length, MENDAKO)


def get_time_string(only_date: bool = False) -> str:
    """
    Return current time as string.
    example
        only_date == True
            '2019-11-08'
        only_date == False
            '2019-11-08 15:12:16.036244'
    :return: time string.
    """
    if only_date:
        return str(datetime.now().date())
    else:
        return str(datetime.now())


def print_with_color(msg: str,
                     color: str) -> None:
    """
    print message to console with color.
    :param msg: message.
    :param color: color, can use data type ConsoleColor.
    :return: None.
    """
    print(color + msg + ConsoleColor.END)


def os_exit(status: int = 0,
            info: str = 'correct') -> None:
    """
    exit system use os._exit.
    :param status: status code.
    :param info: system info.
    :return: None.
    """
    print_with_color(f'System stopped.(OS EXIT)<Code {status}>\nExit info:\n{info}', ConsoleColor.GREEN)
    _exit(status)


def sys_exit(status: int = 0,
             info: str = 'correct') -> None:
    """
    exit system use sys.exit.
    :param status: status code.
    :param info: system info.
    :return: None.
    """
    print_with_color(f'System stopped.(SYS EXIT)<Code {status}>\nExit info:\n{info}', ConsoleColor.GREEN)
    exit(status)


def add_extension(filename: Union[Path, str],
                  extension: str) -> str:
    """
    Adds an extension to filename.
    :param filename: original filename.
    :param extension: extension to add.
    """
    # set name
    name: str = str(filename)
    # set extension
    if extension.startswith('.'):
        ext = extension
    else:
        ext = '.' + extension
    if name.endswith(ext):  # check name
        return name
    else:
        return name + ext


def add_dot_jpg(filename: Union[Path, str]) -> str:
    """Adds a .jpg extension to filename."""
    return add_extension(filename, '.jpg')


def add_dot_jpeg(filename: Union[Path, str]) -> str:
    """Adds a .jpeg extension to filename."""
    return add_extension(filename, '.jpeg')


def add_dot_gif(filename: Union[Path, str]) -> str:
    """Adds a .gif extension to filename."""
    return add_extension(filename, '.gif')


def add_dot_txt(filename: Union[Path, str]) -> str:
    """Adds a .txt extension to filename."""
    return add_extension(filename, '.txt')


def add_dot_png(filename: Union[Path, str]) -> str:
    """Adds a .png extension to filename."""
    return add_extension(filename, '.png')


def add_dot_csv(filename: Union[Path, str]) -> str:
    """Adds a .csv extension to filename."""
    return add_extension(filename, '.csv')


def add_dot_rtf(filename: Union[Path, str]) -> str:
    """Adds a .rtf extension to filename."""
    return add_extension(filename, '.rtf')


def add_dot_pdf(filename: Union[Path, str]) -> str:
    """Adds a .pdf extension to filename."""
    return add_extension(filename, '.pdf')


def add_dot_md(filename: Union[Path, str]) -> str:
    """Adds a .md extension to filename."""
    return add_extension(filename, '.md')


def add_dot_log(filename: Union[Path, str]) -> str:
    """Adds a .log extension to filename."""
    return add_extension(filename, '.log')


def add_dot_json(filename: Union[Path, str]) -> str:
    """Adds a .json extension to filename."""
    return add_extension(filename, '.json')


def add_dot_py(filename: Union[Path, str]) -> str:
    """Adds a .jpeg extension to filename."""
    return add_extension(filename, '.py')


def add_dot_cache(filename: Union[Path, str]) -> str:
    """Adds a .jpeg extension to filename."""
    return add_extension(filename, '.cache')


def add_dot_pickle(filename: Union[Path, str]) -> str:
    """Adds a .pickle extension to filename."""
    return add_extension(filename, '.pickle')


def add_dot_js(filename: Union[Path, str]) -> str:
    """Adds a .js extension to filename."""
    return add_extension(filename, '.js')


def add_dot_css(filename: Union[Path, str]) -> str:
    """Adds a .css extension to filename."""
    return add_extension(filename, '.css')


def add_dot_html(filename: Union[Path, str]) -> str:
    """Adds a .html extension to filename."""
    return add_extension(filename, '.html')


def set_interval(function: Callable,
                 interval: float,
                 count: Optional[int] = None,
                 other_thread: bool = True) -> None:
    """
    Run function (count) times with interval(seconds).
    :param function: target function.
    :param interval: run function interval.
    :param count: run limit. if this value is None, run function without limit.
    :param other_thread: run on other thread.
    :return: None.
    """
    def __inner_wrapper() -> None:
        remaining = count
        while True:
            if (remaining is None) or (remaining > 0):
                try:
                    remaining -= 1
                except TypeError:
                    pass
                sleep(interval)
                function()
            else:
                break
    if other_thread:
        thread = Thread(target=__inner_wrapper)
        thread.start()
    else:
        __inner_wrapper()


def set_timeout(function: Callable,
                timeout: float,
                other_thread: bool = True) -> None:
    """
    Run function with timeout (seconds).
    :param function: target function.
    :param timeout: timeout (seconds).
    :param other_thread: run on other thread.
    :return: None.
    """
    set_interval(function, timeout, count=1, other_thread=other_thread)


def run_on_other_thread(function: Callable,
                        *args,
                        **kwargs) -> Thread:
    """Run function on other thread."""
    thread = Thread(target=function, args=args, kwargs=kwargs)
    thread.start()
    return thread


def on_other_thread(func):
    """This is a decorator, can run function on other thread."""
    def decorator(*args, **kwargs):
        run_on_other_thread(func, *args, **kwargs)
    return decorator


def run_on_other_process(function: Callable,
                         *args,
                         **kwargs) -> Process:
    """Run function on other process."""
    process = Process(target=function, args=args, kwargs=kwargs)
    process.start()
    return process


def on_other_process(func):
    """This is a decorator, can run function on other process."""
    def decorator(*args, **kwargs):
        run_on_other_process(func, *args, **kwargs)
    return decorator


def is_hiragana(text: str) -> bool:
    """Check is text made of hiragana"""
    for character in text:
        if character not in HIRAGANA:
            return False
    return True


def is_small_hiragana(text: str) -> bool:
    """Check is text made of small hiragana"""
    small_hiragana = HIRAGANA[46:55]
    for character in text:
        if character not in small_hiragana:
            return False
    return True


def is_katakana(text: str) -> bool:
    """Check is text made of katakana"""
    for character in text:
        if character not in KATAKANA:
            return False
    return True


def is_small_katakana(text: str) -> bool:
    """Check is text made of small katakana"""
    small_katakana = KATAKANA[46:55]
    for character in text:
        if character not in small_katakana:
            return False
    return True


def hiragana_to_katakana(hiragana: str,
                         hide_other: bool = False) -> str:
    """Convert hiragana characters to katakana characters"""
    if hide_other:
        return ''.join([KATAKANA[HIRAGANA.index(character)] if character in HIRAGANA else
                        character if character in KATAKANA else
                        '' for character in hiragana])
    else:
        return ''.join([KATAKANA[HIRAGANA.index(character)] if character in HIRAGANA else
                        character for character in hiragana])


def katakana_to_hiragana(katakana: str,
                         hide_other: bool = False) -> str:
    """Convert katakana characters to hiragana characters"""
    if hide_other:
        return ''.join([HIRAGANA[KATAKANA.index(character)] if character in KATAKANA else
                        character if character in HIRAGANA else
                        '' for character in katakana])
    else:
        return ''.join([HIRAGANA[KATAKANA.index(character)] if character in KATAKANA else
                        character for character in katakana])


def check_length(min_length: int,
                 max_length: int,
                 mode: str = 'and',
                 *args) -> bool:
    """
    check items length is between min_length and max_length
    :param min_length: minimum length
    :param max_length: maximum length
    :param mode: check mode, 'and': all items need clear length check, 'or': more than one item need clear length check
    :param args: items
    :return: status, [correct] or [incorrect]
    """
    if mode == 'and':
        for item in args:
            if not (min_length <= len(item) <= max_length):
                return False  # if found incorrect item, stop check-loop and return False
        return True  # if can't found incorrect item, return True
    else:
        for item in args:
            if min_length <= len(item) <= max_length:
                return True  # if found correct item, stop check-loop and return True
        return False  # if can't found correct item, return False


def to_string_list(data_list: Iterable) -> List[str]:
    """Convert all item in data_list to string"""
    return list(map(lambda v: str(v), data_list))


def create_path_if_not_exists(path: Union[Path, str]) -> None:
    """Create path if not exists."""
    if isinstance(path, Path):
        target = path
    else:
        target = Path(path)
    if target.is_dir():
        target.mkdir(parents=True, exist_ok=True)
    else:
        target.parent.mkdir(parents=True, exist_ok=True)


def dump_json_beautiful(data: Any,
                        file: Union[StringIO, Path, str]) -> None:
    """Dump json data to file with beautiful format."""
    if isinstance(file, StringIO):
        dump(data, file, ensure_ascii=False, indent=4, sort_keys=True, separators=(',', ': '))
    else:
        create_path_if_not_exists(file)
        with open(str(file), mode='w', encoding='utf-8') as output_file:
            dump(data, output_file, ensure_ascii=False, indent=4, sort_keys=True, separators=(',', ': '))


def dumps_json_beautiful(data: Any) -> str:
    """Dump json data as string with beautiful format."""
    return dumps(data, ensure_ascii=False, indent=4, sort_keys=True, separators=(',', ': '))

# -------------------------------------------------------------------------- Public Functions --
